
#ifndef SOFTAUTH_H
#define SOFTAUTH_H

#include "ownet.h"

#include "D3des.h"
#include "shaib.h"
#include "mbnv.h"
#include "pw77.h"
#include "mbee77.h"

///----------------------------------------------------------------
/// Whether or not DS1977 device is supported
/// #define SA_USE_DS1977
#define SA_DS1977_FAMILY_CODE ((uchar)0x37)
#define SA_MDP_DS1977 (32704)

///----------------------------------------------------------------
/// Whether or not DS1963S device is supported
/// #define SA_USE_DS1963S
#define SA_DS1963S_FAMILY_CODE ((uchar)0x18)
#define SA_MDP_DS1963S (480)

///----------------------------------------------------------------
/// Whether or not DS1961S/DS2432 device is supported
/// #define SA_USE_DS1961S
#define SA_DS1961S_FAMILY_CODE ((uchar)0x33)
#define SA_MDP_DS1961S (128)

///----------------------------------------------------------------
/// Determine max data packet size
#if defined(SA_USE_DS1977)
   #define SA_MAX_DATA_PACKET SA_MDP_DS1977
#elif defined(SA_USE_DS1963S)
   #define SA_MAX_DATA_PACKET SA_MDP_DS1963S
#elif defined(SA_USE_DS1961S)
   #define SA_MAX_DATA_PACKET SA_MDP_DS1961S
#else
   #error "No devices defined as supported.  Must define either SA_USE_DS1977, SA_USE_DS1963S, or SA_USE_DS1961S."
#endif

typedef struct
{
   int hnd;
   uchar k[8];
   uchar romID[8];
   uchar signature[20];
   uchar contents[SA_MAX_DATA_PACKET];
}
DeviceStruct;

///----------------------------------------------------------------
/// Whether or not include code for intiailizing device
/// #define SA_NO_INIT

///----------------------------------------------------------------
/// Whether or not include code for clearing device
/// #define SA_NO_CLEAR

///----------------------------------------------------------------
/// Function Declarations

///
/// Initializes the SA library, opens the 1-Wire port, and finds
/// a 1-Wire device supported by the SA library.   Only returns
/// supported devices, such as the DS1963S, DS1961S/DS2432, and DS1977.
/// Supported devices can be set at compile time by adjusting which of
/// the 'use' flags are defined, such as SA_USE_DS1961S, SA_USE_DS1963S,
/// and SA_USE_DS1977.
///
/// @param szPortStr the hardware port to open for 1-Wire adapter.
/// @return <0 for hardware port failure
///         =0 for no devices found,
///         >0 for success in opening port and finding device
///
int saBeginSession(char* portStr, DeviceStruct* pDevStruct);

///
/// De-initializes the SA library and closes the 1-Wire port.
///
/// @param hnd the handle which refers to the current port.
///
void saEndSession(DeviceStruct* pDevStruct);

///
/// Initializes the device, setting any pins or secret information, and
/// stores the data packet on the device in an encrypted form.
///
/// @param pDevStruct DeviceStruct for holding device information
/// @param szKeyPhrase secure phrase which is used to derive PIN, Secret,
///                    and encryption keys.
/// @param data data packet to store on device
/// @param dataLen length of data packet
/// @return 0 if unsuccessful, >0 otherwise
///
int saInitDevice(DeviceStruct* pDevStruct, char* szKeyPhrase,
                 uchar* data, unsigned int dataLen);

///
/// Authenticates the device, using the derived PIN or SHA secret, and
/// decrypts the data packet stored on the device.
///
/// @param pDevStruct DeviceStruct for holding device information
/// @param szKeyPhrase secure phrase which is used to derive PIN, Secret,
///                    and encryption keys.
/// @param dataLen Expected length of data packet on device
/// @return 0 if unsuccessful, >0 otherwise
///
int saAuthDevice(DeviceStruct* pDevStruct, char* szKeyPhrase,
                 unsigned int dataLen);
///
/// Clears the PIN or secret information from the device, as well as
/// erasing any stored data packets.
///
/// @param pDevStruct DeviceStruct for holding device information
/// @param szKeyPhrase secure phrase which is used to derive PIN, Secret,
///                    and encryption keys.
/// @param dataLen Length of data to erase, 0 for all memory.
/// @return 0 if unsuccessful, >0 otherwise
///
int saClearDevice(DeviceStruct* pDevStruct, char* szKeyPhrase,
                  unsigned int dataLen);

#endif //ifndef SOFTAUTH_H

