// SoftAuth.c
//---------------------------------------------------------------------------
// Copyright (C) 2000 Dallas Semiconductor Corporation, All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Dallas Semiconductor
// shall not be used except as stated in the Dallas Semiconductor
// Branding Policy.
//---------------------------------------------------------------------------
//
// SoftAuth.c - High-level functions for Software Authorization services.
//
// Version: 2.00

#include "SoftAuth.h"

int saBeginSession(char* szPortStr, DeviceStruct* pDevStruct)
{
   int foundDevice = FALSE;
   if((pDevStruct->hnd = owAcquireEx(szPortStr))>=0)
   {
      if(owSpeed(pDevStruct->hnd, MODE_NORMAL) == MODE_NORMAL)
      {
         if(owFirst(pDevStruct->hnd, TRUE, FALSE))
         {
            do
            {
               uchar familyCode;

               owSerialNum(pDevStruct->hnd, pDevStruct->romID, TRUE);
               familyCode = (*pDevStruct->romID)&0x07F;

            #ifdef SA_USE_DS1963S
               if(familyCode == SA_DS1963S_FAMILY_CODE)
                  foundDevice = TRUE;
            #endif
            #ifdef SA_USE_DS1961S
               if(familyCode == SA_DS1961S_FAMILY_CODE)
                  foundDevice = TRUE;
            #endif
            #ifdef SA_USE_DS1977
               if(familyCode == SA_DS1977_FAMILY_CODE)
                  foundDevice = TRUE;
            #endif

            }
            while(!foundDevice && owNext(pDevStruct->hnd, TRUE, FALSE));

            // put device in overdrive mode
            (void)owTouchReset(pDevStruct->hnd);
            (void)owTouchByte(pDevStruct->hnd, 0x3C);
            (void)owSpeed(pDevStruct->hnd, MODE_OVERDRIVE);
            (void)owTouchReset(pDevStruct->hnd);
         }
      }
      else
         return -2;
   }
   else
      return -1;
   return foundDevice;
}

void saEndSession(DeviceStruct* pDevStruct)
{
   owRelease(pDevStruct->hnd);
   pDevStruct->hnd = -1;
}

int saInitDevice(DeviceStruct* pDevStruct, char* szKeyPhrase, uchar* data, unsigned int dataLen)
{
#ifndef SA_NO_INIT
   uchar familyCode = (uchar)(*pDevStruct->romID)&0x07F;
   uchar k[24], u1[8], u2[8];
   unsigned int i;
   char* szPhraseCpy;
   size_t phraseLen = strlen(szKeyPhrase);
   unsigned int dataLenPrime = dataLen + (8-dataLen%8);

   if(dataLenPrime>SA_MAX_DATA_PACKET)
      return FALSE;

   szPhraseCpy = malloc(phraseLen+16+1);
   if(szPhraseCpy==NULL)
      return FALSE;

   strcpy(szPhraseCpy, szKeyPhrase);
   sprintf(&szPhraseCpy[phraseLen],
      "%02X%02X%02X%02X%02X%02X%02X%02X",
      pDevStruct->romID[0], pDevStruct->romID[1], pDevStruct->romID[2], pDevStruct->romID[3],
      pDevStruct->romID[4], pDevStruct->romID[5], pDevStruct->romID[6], pDevStruct->romID[7]);
   szPhraseCpy[phraseLen+16] = 0;

   memset(k, 0, sizeof(k));
   make3key(szPhraseCpy, k);
   des3key(k, EN0);

   free(szPhraseCpy);

   for(i=0; i<dataLen; i+=8)
   {
      memset(u1, 0, 8);
      if(i+8>dataLen)
         memcpy(u1, &data[i], (size_t)dataLen-i);
      else
         memcpy(u1, &data[i], (size_t)8);
      memcpy(u2, u1, 8);
      Ddes(u1, &pDevStruct->contents[i]);
   }

#ifdef SA_USE_DS1977
   if(familyCode == SA_DS1977_FAMILY_CODE)
   {
      if(dataLen>SA_MDP_DS1977)
         return FALSE;

      Ddes(pDevStruct->romID, u1);
      if(setPassword(pDevStruct->hnd, pDevStruct->romID, u1, 1))
      {
         setBMPassword(u1);
         if(setPasswordMode(pDevStruct->hnd, pDevStruct->romID, 1))
         {
            return writeEE77(1, pDevStruct->hnd, pDevStruct->romID,
                             0, pDevStruct->contents, dataLenPrime);
         }
      }
   }
#endif
#ifdef SA_USE_DS1961S
   if(familyCode == SA_DS1961S_FAMILY_CODE)
   {
      uchar fullBindCode[15];
      uchar skey[8];
      SHACopr copr = {0};
      SHAUser user = {0};
      size_t inputMsgLen = phraseLen + 8;
      uchar* inputMsg = malloc(inputMsgLen);

      if(inputMsg==NULL)
         return FALSE;

      if(dataLen>SA_MDP_DS1961S)
      {
         free(inputMsg);
         return FALSE;
      }

      memset(copr.bindCode, 0xAA, 7);
      memset(copr.bindData, 0x55, 32);

      memcpy(user.devAN, pDevStruct->romID, 8);
      user.portnum = pDevStruct->hnd;
      user.accountPageNumber = (uchar)0;

      memset(inputMsg, 0x0FF, inputMsgLen);
      memset(skey, 0, sizeof(skey));
      Ddes(pDevStruct->romID, skey);
      memset(u1, 0, sizeof(u1));
      memset(u2, 0, sizeof(u2));
      memcpy(inputMsg, skey, 8);
      for(i = 0; i<(unsigned int)phraseLen; i += 8)
      {
         size_t rem = min(phraseLen - i, 8);
         memset(u1, 0x0FF, 8);
         memcpy(u1, &szKeyPhrase[i], rem);
         Ddes(u1, u2);
         memcpy(&inputMsg[8+i], u2, rem);
      }

      if(!InstallSystemSecret33(pDevStruct->hnd,
                                (int)user.accountPageNumber,
                                (int)user.accountPageNumber&0x07,
                                inputMsg, (int)inputMsgLen, FALSE))
      {
         free(inputMsg);
         OWERROR(OWERROR_INSTALL_SECRET_FAILED);
         return FALSE;
      }

      memcpy(fullBindCode, copr.bindCode, 4);
      fullBindCode[4] = (uchar)user.accountPageNumber;
      memcpy(&fullBindCode[5], user.devAN, 7);
      memcpy(&fullBindCode[12], &(copr.bindCode[4]), 3);

      if( !BindSecretToiButton33(user.portnum,
                                 (int)user.accountPageNumber,
                                 (int)user.accountPageNumber&7,
                                 copr.bindData, fullBindCode, TRUE))
      {
         free(inputMsg);
         OWERROR(OWERROR_BIND_SECRET_FAILED);
         return FALSE;
      }

      (void)InstallAuthSecretVM(&copr, inputMsg, (int)inputMsgLen);

      free(inputMsg);

      if(VerifyUser(&copr, &user, TRUE))
      {
         SMALLINT success = TRUE;
         unsigned int pages = dataLenPrime/32 + (dataLenPrime%32?1:0);
         for(i=0; success && i<4 && i<pages; i++)
         {
            user.accountPageNumber = (uchar)i;
            memcpy(user.accountFile, &pDevStruct->contents[i*32],
               (size_t)(min(32, dataLenPrime-i*32)));
            success = WriteDataPageSHA33(&copr, &user);
         }
         return success;
      }
   }
#endif
#ifdef SA_USE_DS1963S
   if(familyCode == SA_DS1963S_FAMILY_CODE)
   {
      uchar fullBindCode[15];
      uchar skey[8];
      SHACopr copr = {0};
      SHAUser user = {0};
      size_t inputMsgLen = phraseLen + 8;
      uchar* inputMsg = malloc(inputMsgLen);

      if(inputMsg==NULL)
         return FALSE;

      if(dataLen>SA_MDP_DS1963S)
      {
         free(inputMsg);
         return FALSE;
      }

      memset(copr.bindCode, 0xAA, 7);
      memset(copr.bindData, 0x55, 32);

      memcpy(user.devAN, pDevStruct->romID, 8);
      user.portnum = pDevStruct->hnd;
      user.accountPageNumber = (uchar)15;

      memset(inputMsg, 0x0FF, inputMsgLen);
      memset(skey, 0, sizeof(skey));
      Ddes(pDevStruct->romID, skey);
      memset(u1, 0, sizeof(u1));
      memset(u2, 0, sizeof(u2));
      memcpy(inputMsg, skey, 8);
      for(i = 0; i<(unsigned int)phraseLen; i += 8)
      {
         size_t rem = min((phraseLen) - ((size_t)i), 8);
         memset(u1, 0x0FF, 8);
         memcpy(u1, &szKeyPhrase[i], rem);
         Ddes(u1, u2);
         memcpy(&inputMsg[8+i], u2, rem);
      }

      if(!InstallSystemSecret18(pDevStruct->hnd,
                                (int)user.accountPageNumber,
                                (int)user.accountPageNumber&0x07,
                                inputMsg, (int)inputMsgLen, FALSE))
      {
         free(inputMsg);
         OWERROR(OWERROR_INSTALL_SECRET_FAILED);
         return FALSE;
      }

      memcpy(fullBindCode, copr.bindCode, 4);
      fullBindCode[4] = (uchar)user.accountPageNumber;
      memcpy(&fullBindCode[5], user.devAN, 7);
      memcpy(&fullBindCode[12], &(copr.bindCode[4]), 3);

      if(!BindSecretToiButton18(user.portnum,
                                (int)user.accountPageNumber,
                                (int)user.accountPageNumber&7,
                                copr.bindData, fullBindCode, TRUE))
      {
         free(inputMsg);
         OWERROR(OWERROR_INSTALL_SECRET_FAILED);
         return FALSE;
      }

      (void)InstallAuthSecretVM(&copr, inputMsg, (int)inputMsgLen);

      free(inputMsg);

      if(VerifyUser(&copr, &user, TRUE))
      {
         uchar buff[32];
         uchar scratch[32];
         uchar mac[20];
         uchar wccTemp[32];
         int inc;
         unsigned __int64 wccSum = 0;

         memset(wccTemp, 0, sizeof(wccTemp));
         if(readNV(3, user.portnum, user.devAN, 0, FALSE, wccTemp, 32))
         {
            unsigned int lastWcc = 0;
            for(i = 0; i<8; i++)
            {
               lastWcc = (unsigned int)((wccTemp[i*4+0]<<0)  |
                               (wccTemp[i*4+1]<<8)  |
                               (wccTemp[i*4+2]<<16) |
                               (wccTemp[i*4+3]<<24));
               wccSum += lastWcc;
            }

            inc = (dataLen/32 + (dataLen%32?1:0) - 8);
            if(inc>0)
               wccSum += inc;
            wccSum += 1;
            lastWcc += 1;

            memset(mac, 0, sizeof(mac));
            memset(scratch, 0, 32);
            for(i=0; i<dataLenPrime; i+=32)
            {
               size_t rem = (size_t)(min(((int)dataLenPrime)-((int)i), 32));
               memset(buff, 0, 32);
               memcpy(buff, &pDevStruct->contents[i], rem);
               (void)CreateDataSignatureVM(&copr, skey, buff, scratch, mac, TRUE);
               memcpy(&scratch[8], mac, 20);
               OWASSERT( WriteDataPageSHA18(user.portnum, (int)i/32, buff, FALSE),
                        OWERROR_WRITE_DATA_PAGE_FAILED, FALSE );
            }

            memset(buff, 0, 32);
            for(i=0; i<5; i++)
            {
               buff[i] = (uchar)(wccSum&0x0FF);
               wccSum = wccSum>>8;
            }
            for(i=5; i<9; i++)
            {
               buff[i] = (uchar)(lastWcc&0x0FF);
               lastWcc = lastWcc>>8;
            }
            (void)CreateDataSignatureVM(&copr, skey, buff, scratch, mac, TRUE);
            memcpy(&buff[9], mac, 20);
            OWASSERT( WriteDataPageSHA18(user.portnum, 15, buff, FALSE),
                     OWERROR_WRITE_DATA_PAGE_FAILED, FALSE );
            return TRUE;
         }
      }
   }
#endif
#endif // ifndef SA_NO_INIT
   return FALSE;
}

int saAuthDevice(DeviceStruct* pDevStruct, char* szKeyPhrase, unsigned int dataLen)
{
   uchar familyCode = (*pDevStruct->romID)&0x07F;
   unsigned int i;
   uchar k[24], u1[8], u2[8];
   char* szPhraseCpy;
   size_t phraseLen = strlen(szKeyPhrase);
   unsigned int dataLenPrime = dataLen + (8-dataLen%8);

   szPhraseCpy = malloc(phraseLen+16+1);
   if(szPhraseCpy==NULL)
      return FALSE;

   strcpy(szPhraseCpy, szKeyPhrase);
   sprintf(&szPhraseCpy[phraseLen],
      "%02X%02X%02X%02X%02X%02X%02X%02X",
      pDevStruct->romID[0], pDevStruct->romID[1], pDevStruct->romID[2], pDevStruct->romID[3],
      pDevStruct->romID[4], pDevStruct->romID[5], pDevStruct->romID[6], pDevStruct->romID[7]);
   szPhraseCpy[phraseLen+16] = 0;

   memset(k, 0, sizeof(k));
   make3key(szPhraseCpy, k);
   des3key(k, EN0);

   free(szPhraseCpy);

   if(dataLenPrime>SA_MAX_DATA_PACKET)
      return FALSE;

#ifdef SA_USE_DS1977
   if(familyCode == SA_DS1977_FAMILY_CODE)
   {
      Ddes(pDevStruct->romID, u1);
      if(verifyPassword(pDevStruct->hnd, pDevStruct->romID, u1, 1))
      {
         setBMPassword(u1);
         readEEPsw77(0, pDevStruct->hnd, pDevStruct->romID, 0, FALSE, pDevStruct->contents, dataLenPrime);
      }
   }
#endif
#ifdef SA_USE_DS1961S
   if(familyCode == SA_DS1961S_FAMILY_CODE)
   {
      uchar skey[8];
      SHACopr copr;
      SHAUser user;
      size_t inputMsgLen = phraseLen + 8;
      uchar* inputMsg = malloc(inputMsgLen);

      if(inputMsg==NULL)
         return FALSE;

      if(dataLen>SA_MDP_DS1961S)
      {
         free(inputMsg);
         return FALSE;
      }

      memset(copr.bindCode, 0xAA, 7);
      memset(copr.bindData, 0x55, 32);

      memcpy(user.devAN, pDevStruct->romID, 8);
      user.portnum = pDevStruct->hnd;
      user.accountPageNumber = (uchar)0;

      memset(inputMsg, 0x0FF, inputMsgLen);
      memset(skey, 0, sizeof(skey));
      Ddes(pDevStruct->romID, skey);
      memset(u1, 0, sizeof(u1));
      memset(u2, 0, sizeof(u2));
      memcpy(inputMsg, skey, 8);
      for(i = 0; i<(unsigned int)phraseLen; i += 8)
      {
         size_t rem = min(phraseLen - i, 8);
         memset(u1, 0x0FF, 8);
         memcpy(u1, &szKeyPhrase[i], rem);
         Ddes(u1, u2);
         memcpy(&inputMsg[8+i], u2, rem);
      }

      (void)InstallAuthSecretVM(&copr, inputMsg, (int)inputMsgLen);

      free(inputMsg);

      for(i=0; i<4 && i<(dataLenPrime/32+(dataLenPrime%32?1:0)); i++)
      {
         user.accountPageNumber = (uchar)i;
         if(VerifyUser(&copr, &user, i==0))
         {
            memcpy(&pDevStruct->contents[i*32], user.accountFile,
               (size_t)(min(32, dataLenPrime-i*32)));
         }
      }
   }
#endif
#ifdef SA_USE_DS1963S
   if(familyCode == SA_DS1963S_FAMILY_CODE)
   {
      uchar contentBuff[SA_MDP_DS1963S];
      uchar skey[8];
      SHACopr copr;
      SHAUser user;
      uchar scratch[32];
      uchar mac[20];
      unsigned __int64 wccSum = 0, wccSumTemp = 0;
      unsigned int lastWcc = 0;

      size_t inputMsgLen = phraseLen + 8;
      uchar* inputMsg = malloc(inputMsgLen);

      if(inputMsg==NULL)
         return FALSE;

      if(dataLen>SA_MDP_DS1963S)
      {
         free(inputMsg);
         return FALSE;
      }

      memset(contentBuff, 0, sizeof(contentBuff));

      memset(copr.bindCode, 0xAA, 7);
      memset(copr.bindData, 0x55, 32);

      memcpy(user.devAN, pDevStruct->romID, 8);
      user.portnum = pDevStruct->hnd;
      user.accountPageNumber = (uchar)15;

      memset(inputMsg, 0x0FF, inputMsgLen);
      memset(skey, 0, sizeof(skey));
      Ddes(pDevStruct->romID, skey);
      memset(u1, 0, sizeof(u1));
      memset(u2, 0, sizeof(u2));
      memcpy(inputMsg, skey, 8);
      for(i = 0; i<(unsigned int)phraseLen; i += 8)
      {
         size_t rem = min(phraseLen - i, 8);
         memset(u1, 0x0FF, 8);
         memcpy(u1, &szKeyPhrase[i], rem);
         Ddes(u1, u2);
         memcpy(&inputMsg[8+i], u2, rem);
      }

      (void)InstallAuthSecretVM(&copr, inputMsg, (int)inputMsgLen);

      free(inputMsg);

      memset(mac, 0, sizeof(mac));
      memset(scratch, 0, 32);
      for(i=0; i<15; i++)
      {
         int rem = (dataLenPrime)-(i*32);
         unsigned int wcc = (unsigned int)ReadAuthPageSHA18(user.portnum, (int)i, user.accountFile,
                                            NULL, FALSE);
         if(i>7)
            wccSum += wcc;
         if(rem>0)
         {
            (void)CreateDataSignatureVM(&copr, skey, user.accountFile,
                                 scratch, mac, TRUE);
            memcpy(&contentBuff[i*32], user.accountFile, (size_t)rem);
            memcpy(&scratch[8], mac, 20);
         }
      }
      if(VerifyUser(&copr, &user, TRUE))
      {
         wccSum += user.writeCycleCounter;
         for(i=0; i<5; i++)
         {
            wccSumTemp = wccSumTemp << 8;
            wccSumTemp = wccSumTemp | user.accountFile[4-i];
         }
         for(i=0; i<4; i++)
         {
            lastWcc = lastWcc << 8;
            lastWcc = lastWcc | user.accountFile[8-i];
         }
         if(wccSum==wccSumTemp && lastWcc==(unsigned int)user.writeCycleCounter)
         {
            uchar buff[32];
            memset(buff, 0, 32);
            memcpy(buff, user.accountFile, 9);
            (void)CreateDataSignatureVM(&copr, skey, buff,
                                    scratch, mac, TRUE);
            if(memcmp(&user.accountFile[9], mac, 20)==0)
            {
               memcpy(pDevStruct->contents, contentBuff, (size_t)dataLenPrime);
            }
         }
      }
   }
#endif

   des3key(k, DE1);
   for(i=0; i<dataLenPrime; i+=8)
   {
      memcpy(u2, &pDevStruct->contents[i], 8);
      Ddes(u2, &pDevStruct->contents[i]);
   }

   return TRUE;
}

int saClearDevice(DeviceStruct* pDevStruct, char* szKeyPhrase, unsigned int dataLen)
{
#ifndef SA_NO_CLEAR
   unsigned int i;
   uchar k[24], u1[8], u2[8];
   char* szPhraseCpy;
   size_t phraseLen = strlen(szKeyPhrase);

   szPhraseCpy = malloc(phraseLen+16+1);
   if(szPhraseCpy==NULL)
      return FALSE;

   strcpy(szPhraseCpy, szKeyPhrase);
   sprintf(&szPhraseCpy[phraseLen],
      "%02X%02X%02X%02X%02X%02X%02X%02X",
      pDevStruct->romID[0], pDevStruct->romID[1], pDevStruct->romID[2], pDevStruct->romID[3],
      pDevStruct->romID[4], pDevStruct->romID[5], pDevStruct->romID[6], pDevStruct->romID[7]);
   szPhraseCpy[phraseLen+16] = 0;

   memset(k, 0, sizeof(k));
   make3key(szPhraseCpy, k);
   des3key(k, EN0);

   free(szPhraseCpy);

   memset(u1, 0, sizeof(u1));
   memset(u2, 0, sizeof(u2));

#ifdef SA_USE_DS1977
   if(*pDevStruct->romID == SA_DS1977_FAMILY_CODE)
   {
      Ddes(pDevStruct->romID, u1);
	  
      setBMPassword(u1);
      if(setPasswordMode(pDevStruct->hnd, pDevStruct->romID, 0))
      {
         if(setPassword(pDevStruct->hnd, pDevStruct->romID, u2, 1))
         {
			unsigned int overwriteLen = SA_MDP_DS1977;
			if(dataLen>0 && dataLen<SA_MDP_DS1977)
		       overwriteLen = dataLen;

            memset(pDevStruct->contents, 0xAA, overwriteLen);
            return writeEE77(1, pDevStruct->hnd, pDevStruct->romID,
                             0, pDevStruct->contents, overwriteLen);
         }
      }
   }
#endif
#ifdef SA_USE_DS1961S
   if(*pDevStruct->romID == SA_DS1961S_FAMILY_CODE)
   {

      SHAUser user = {0};
      SHACopr copr = {0};
      uchar fullBindCode[15];
      uchar inputMsg[32];
      unsigned int overwriteLen = SA_MDP_DS1961S;
	  if(dataLen>0 && dataLen<SA_MDP_DS1961S)
	     overwriteLen = dataLen;

      memset(inputMsg, 0xAA, 32);

      memset(copr.bindCode, 0xAA, 7);
      memset(copr.bindData, 0x55, 32);

      memcpy(user.devAN, pDevStruct->romID, 8);
      user.portnum = pDevStruct->hnd;
      user.accountPageNumber = (uchar)0;

      memcpy(fullBindCode, copr.bindCode, 4);
      fullBindCode[4] = (uchar)user.accountPageNumber;
      memcpy(&fullBindCode[5], user.devAN, 7);
      memcpy(&fullBindCode[12], &(copr.bindCode[4]), 3);

      memcpy(user.devAN, pDevStruct->romID, 8);
      user.portnum = pDevStruct->hnd;
      user.accountPageNumber = (uchar)0;

      OWASSERT(
         InstallSystemSecret33(user.portnum,
                                (int)0, (int)0,
                                inputMsg, 32, FALSE),
         OWERROR_INSTALL_SECRET_FAILED, FALSE);

      OWASSERT(
         BindSecretToiButton33(user.portnum,
                                (int)0, (int)0,
                                copr.bindData, fullBindCode, TRUE),
         OWERROR_INSTALL_SECRET_FAILED, FALSE);

      (void)InstallAuthSecretVM(&copr, inputMsg, 32);

      OWASSERT(
         VerifyUser(&copr, &user, TRUE),
         OWERROR_INSTALL_SECRET_FAILED, FALSE);

      for(i=0; i<(overwriteLen/32 + ((overwriteLen%32)?1:0)); i++)
      {
         user.accountPageNumber = i;
         memcpy(user.accountFile, inputMsg, 32);
         OWASSERT(
            WriteDataPageSHA33(&copr, &user),
            OWERROR_WRITE_DATA_PAGE_FAILED, FALSE);
      }

      return TRUE;
   }
#endif
#ifdef SA_USE_DS1963S
   if(*pDevStruct->romID == SA_DS1963S_FAMILY_CODE)
   {
      uchar inputMsg[32];
      unsigned int overwriteLen = SA_MDP_DS1963S;
	  if(dataLen>0 && dataLen<SA_MDP_DS1963S)
	     overwriteLen = dataLen;

      memset(inputMsg, 0xAA, 32);
      for(i=0; i<(overwriteLen/32 + ((overwriteLen%32)?1:0)); i++)
      {
         OWASSERT(
            InstallSystemSecret18(pDevStruct->hnd,
                                  (int)i, (int)i&0x07,
                                  inputMsg, 32, FALSE),
            OWERROR_INSTALL_SECRET_FAILED, FALSE);
      }

      return TRUE;
   }
#endif
#endif // ifndef SA_NO_CLEAR

   return FALSE;
}
